"""
Class for configuration of the pipeline, models and experiments for better tracking in one place.
"""

class ExperimentConfig:
    pass


class PipelineConfig:
    pass


class ModelConfig:
    """
    Example:

    kwargs = {
        "model_config": {
            "pretrained_model_name_or_path": model_name,
            "device": "cpu",
            # "device_map": "auto", # put to GPU if GPU is available
            "max_position_embeddings": MAX_LENGTH,
            "max_length": MAX_LENGTH,
        },
    }
    tokenizer_config = ModelConfig(**kwargs)
    tokenizer = AutoTokenizer.from_pretrained(
        **tokenizer_config.get_config(), 
        **token_kwargs,
    )
    """
    def __init__(self, model_config: dict = {}, **kwargs):
        self.model_config = model_config

    def get_config(self) -> dict:
        return self.model_config

    def __repr__(self):
        # add !r to format the string representation of the object
        return f"ModelConfig(model_config={self.model_config!r})"
    
    def __str__(self):
        return repr(self)
    

class LocalCausalLMConfig:
    """
    Example:
    base_lm_config = ModelConfig(
        model_config = {
            "pretrained_model_name_or_path": model_name,
            "device_map": "auto",
            # "max_memory": f"{int(torch.cuda.mem_get_info()[0]/1024**3)-2}GB",
        }
    )
    kwargs = {
        "quantized": True,
        "model_config": base_lm_config.get_config(),
        "quantization_config": {
            "quantization_config": transformers.BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type='nf4',
                bnb_4bit_use_double_quant=True,
                bnb_4bit_compute_dtype=bfloat16
            )
        }
    }

    lm_config = LocalCausalLMConfig(**kwargs)

    model = AutoModelForCausalLM.from_pretrained(
      **lm_config.get_config(),
      **token_kwargs,  
    )
    """
    def __init__(self, quantized: bool = False, model_config: dict = {},  quantization_config: dict = {}, **kwargs):
        self.quantized = quantized
        self.model_config = model_config
        self.quantization_config = quantization_config

    def get_config(self) -> dict:
        if self.quantized:
            return {**self.model_config, **self.quantization_config}
        else:
            return self.model_config

    def __repr__(self):
        # add !r to format the string representation of the object
        return f"LocalCausalLMConfig(quantized={self.quantized!r}, model_config={self.model_config!r}, quantization_config={self.quantization_config!r})"
    
    def __str__(self):
        return repr(self)


from time import time
def time_func(f: callable):
    """
    decorator to time a function duration.

    Usage:
    @time_func
    def my_function():
        pass
    
    my_function()

    Args:
        f (callable): function to time, which can be a generator function, async function, or normal function.
    
    Returns:
        callable: decorated function with time duration printed.
    """
    def inner(*args, **kwargs):
        start = time()
        r = f(*args, **kwargs)
        end = time()
        duration = end - start
        # Get name of callable:
        # https://stackoverflow.com/questions/33576932/do-all-callables-have-name/33578357#33578357
        print("="*20)
        print(f"executed: {getattr(f, '__name__', repr(f))}() python function")
        print(f"walltime: {duration} in secs.")
        print("="*20)
        return r
    return inner


import sys, os, glob
from pypdf import PdfReader
from typing import Tuple, List
from langchain_core.documents.base import Document


"""
https://pythonhosted.org/multidispatch/
"""
class PDFHelper():
    def __init__(self, data_folder: str, file_pattern: str):
        """
        Args:
          data_folder: where all the pdf files are
          file_pattern: the pdf files match this pattern
          
        Examples:
           
          PDFHelper(data_folder = "./data/medreports", 
                    file_pattern="KK-SCIVIAS-*.pdf")
        """
        self.data_folder = data_folder
        self.file_pattern = file_pattern
        self.dir_path = f"{data_folder}/{file_pattern}"
        self.file_path_list = glob.glob(self.dir_path)
        
    # @multifunction(None, str)
    def read_pdf(self, input) -> str:
        """read from the give path the text and returns a raw string. 
           use print to print the content
        """
        if isinstance(input, str):
            file_path = input
            reader = PdfReader(file_path)
            content_raw_str = "".join([page.extract_text() for page in reader.pages])
            return content_raw_str
        elif isinstance(input, int):
            file_idx = input
            if (file_idx < len(self.file_path_list)):
                return self.read_pdf(str(self.file_path_list[file_idx]))
            else:
                return ""
        else:
            return ""
    
    # @multifunction(None, str)
    def count_token(self, input)-> int:
        """count the total token in a pdf file
        """
        if isinstance(input, str):
            file_path = input
            token_size = len(self.read_pdf(file_path))
            print(f"file: {file_path}\n" + 
              f"total token: {token_size}")
            return token_size
        elif isinstance(input, int):
            file_idx = input
            if (file_idx < len(self.file_path_list)):
                return self.count_token(str(self.file_path_list[file_idx]))
            else:
                return 0
        else:
            return 0
              
    def read_txt(self, input) -> str:
        """read from the give path the text and returns a raw string. 
           use print to print the content
        """
        if isinstance(input, str):
            file_path = input
            with open(file_path, "r") as txt_file:
                content_raw_str = txt_file.read()
            return content_raw_str
        elif isinstance(input, int):
            file_idx = input
            if (file_idx < len(self.file_path_list)):
                return self.read_txt(str(self.file_path_list[file_idx]))
            else:
                return ""
        else:
            return ""
        

class DocMetaInfo():
    def __init__(self, doc: Document):
        self.read_meta(doc)
    
    
    def read_meta(self, doc: Document):
        file_content = doc.page_content
        self.source = doc.metadata['source']
        self.name = self.source.split("/")[-1]
        self.token_size = len(file_content.split())
        self.character_size = len(file_content)
        
        
    def __str__(self):
        """call by print"""
        return f"source:{self.source}\nname:{self.name}\ntokens:{self.token_size}\ncharacters:{self.character_size}"
    
    
    def __repr__(self):
        """convert obj to string, called by jupyter cell"""
        return self.__str__()


class DocCorpusS3(): 
    def __init__(self, data: List[Document]):
        # enumerate returns a key, element tuple, the x[1] is the DocMetaInfo(doc)
        # https://stackoverflow.com/questions/16945518/finding-the-index-of-the-value-which-is-the-min-or-max-in-python/16945868#16945868
        self.docs = data
        self.docs_meta_infos = [DocMetaInfo(doc) for doc in data]
        self.idx_of_max_token, self.max_doc_meta = max(enumerate(self.docs_meta_infos), key=lambda x: x[1].token_size)
        self.idx_of_min_token, self.min_doc_meta = min(enumerate(self.docs_meta_infos), key=lambda x: x[1].token_size)
    

    def get_s3_obj_info(self, idx: int, show_content: bool = False) -> Tuple[Document, DocMetaInfo]:
        if (self.docs is not None):
            n = len(self.docs)
            print(f"total objects: {n}")
            print("="*20)
            if -n <= idx < n: # in range of list idx
                doc = self.docs[idx]
                meta_info = self.docs_meta_infos[idx]
                file_content = doc.page_content
                
                print(f"s3 key     :{meta_info.source}")
                print(f"obj name   :{meta_info.name}")
                print(f"token size :{meta_info.token_size}")
                print(f"char. size :{meta_info.character_size}")
                if show_content:
                    print("-"*20)
                    print(file_content)
            return doc, meta_info
        else:
            return None


"""
Run this script to download a T5 family translator model

Example: 
>> python3 download_translator.py --help 
"""
import os
import click

'''set the model download cache directory'''
# set the home directory for the models

from dataclasses import dataclass
@dataclass
class DirectorySetting:
    """set the directory for the model download"""
    home_dir: str="/home/jovyan/llm-models"
    transformers_cache_home: str="core-kind/yinwang/models"
    huggingface_token_file: str="core-kind/yinwang/.cache/huggingface/token"

    def get_cache_home(self):
        """get the cache home"""
        return f"{self.home_dir}/{self.transformers_cache_home}"
    
    def get_token_file(self):
        """get the token file"""
        return f"{self.home_dir}/{self.huggingface_token_file}"


# from transformers import T5Tokenizer, T5ForConditionalGeneration

# T5 model family info on huggingface
# https://huggingface.co/docs/transformers/model_doc/t5
# Note: T5 can only translate from "translate_en_to_de" but not de_to_en
# need to use mt5 

# model_map = {
#    "small": "t5-small",
#    "base" : "t5-base",
#    "large" : "t5-large",
#    "3B" : "t5-3b", # 11.4 GB
#    "11B" : "t5-11b" # 45.2 GB
# }

model_map = {
   "small": "google/mt5-small", # 1.2 GB
   "base" : "google/mt5-base", # 2.33 GB
   "large" : "google/mt5-large", # 4.9 GB,
   "xl" : "google/mt5-xl", # 15 GB
   "xxl" : "google/mt5-xxl" # 51.7 GB
}

dir_mode_map = {
    "kf_notebook": DirectorySetting(),
    "mac_local": DirectorySetting(home_dir="/Users/yingding", transformers_cache_home="MODELS", huggingface_token_file="MODELS/.huggingface_token"),
}

default_model_type = "small"
default_dir_mode = "kf_notebook"


def need_token(model_type: str, model_name_prefix: str="llama"):
    """check if the model needs token"""
    return model_type.startswith(model_name_prefix)


def get_token(dir_setting: DirectorySetting):
    """get the token from the token file"""
    token_file_path = dir_setting.get_token_file()
    with open(token_file_path, "r") as file:
        # file read add a new line to the token, remove it.
        token = file.read().replace('\n', '')
    return token

# token_file_path = f"{DATA_ROOT}/core-kind/yinwang/.cache/huggingface/token"
# file = open(token_file_path, "r")

# file read add a new line to the token, remove it.
# token = file.read().replace('\n', '')

# print the raw string to see if there is new line in the token
# print(r'{}'.format(token))


# Reference: https://click.palletsprojects.com/en/8.1.x/quickstart/
# @click.option(..., is_flag=True, ...) set the option to be boolean
# call with download_llms --help
# https://www.geeksforgeeks.org/argparse-vs-docopt-vs-click-comparing-python-command-line-parsing-libraries/
# https://click.palletsprojects.com/en/8.1.x/quickstart/
# https://click.palletsprojects.com/en/8.1.x/options/
# https://www.youtube.com/watch?v=kNke39OZ2k0
@click.command()
@click.option('-t','--model-type', 'model_type', default=default_model_type, type=str, required=False, help=f"set the translator type to download:\n{', '.join(model_map.keys())}, default is {default_model_type}")
@click.option('-m','--mode', 'dir_mode', default=default_dir_mode, type=str, required=False, help=f"set the directory settings to use:\n{', '.join(dir_mode_map.keys())}, default is {default_dir_mode}")
def download(model_type: str=default_model_type, dir_mode: str=default_dir_mode):
    """
    This method will download a T5 translator model. If cache exists, the cached model will be used.

    get help:
    python3 download_translator.py --help

    valid call:
    python3 download_translator.py -t small
    python3 download_translator.py --model-type small
    
    
    set directory:
    python3 download_translator.py -t small -m kf_notebook
    
    Args:
      model_type: "small", ..., "3B", "11B"
    """
    # model_type = params.get("model-type", os.environ['model-type'])
    dir_setting=dir_mode_map.get(dir_mode, dir_mode_map[default_dir_mode])
    os.environ['XDG_CACHE_HOME']=dir_setting.get_cache_home()
    print("-"*10)
    print(f"download dir: {os.environ['XDG_CACHE_HOME']}")
    
    from transformers import MT5Model, MT5ForConditionalGeneration, MT5TokenizerFast, MT5Config

    # model_name = model_map.get(model_type, model_map[os.environ['model-type']])
    model_name = model_map.get(model_type, model_map[default_model_type])
    
    print("-"*10)
    print(f"model_type: {model_type}")
    print(f"model_name: {model_name}")
    print("-"*10)

    if need_token(model_type):
        # kwargs = {"use_auth_token": get_token(dir_setting)}
        kwargs = {"token": get_token(dir_setting)}
        print("huggingface token loaded")
    else:
        kwargs = {}
        print("huggingface token is NOT needed")
    print("-"*10)
    
    tokenizer = MT5TokenizerFast.from_pretrained(model_name)
    model = MT5ForConditionalGeneration.from_pretrained(model_name)
    
    # tokenizer = T5Tokenizer.from_pretrained(model_name)
    # model = T5ForConditionalGeneration.from_pretrained(model_name)
    
    
if __name__ == '__main__':
    """
    uses default for click
    https://stackoverflow.com/questions/49011223/python-correct-use-of-click-with-main-and-init
    
    translator on huggingface: 
    https://huggingface.co/learn/nlp-course/chapter7/4?fw=tf
    
    Inference and Fine-tuning T5 model
    https://huggingface.co/docs/transformers/model_doc/t5
    """
    download()


from .objectstore_utils import (
    S3AccessConf, S3BucketHelper, S3PdfObjHelper, 
)

from .pdf_text_loader import (
    PDFHelper, 
    DocMetaInfo,
    DocCorpusS3
)

from .sqldbs_utils import (
    SqlDBConfig, SqlDBHelperFactory
)

__all__ = [S3AccessConf, S3BucketHelper, S3PdfObjHelper, PDFHelper, DocMetaInfo, DocCorpusS3, SqlDBConfig, SqlDBHelperFactory]


import boto3
from dataclasses import dataclass
from operator import attrgetter
from io import BytesIO
#from boto3.resources.factory.s3 import ServiceResource
from pypdf import PdfReader


@dataclass
class S3AccessConf():
    """
    Examples:
        s3_conf = S3AccessConf(
            bucket_name = "xxx",
            access_key_id = os.environ.get('AWS_ACCESS_KEY_ID'),
            secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY'),
            endpoint = os.environ.get('S3_ENDPOINT'),
            verify_host = True
        )    
    """
    access_key_id: str
    secret_access_key: str
    endpoint: str
    bucket_name: str
    verify_host: bool = True
    
    
class S3BucketHelper():
    # more about client caching
    def __init__(self, conf: S3AccessConf, file_prefix: str=""):
        """
        Args:
            file_prefix: s3 bucket objects key prefix used to filter the object
          
        Examples:
            S3BucketHelper(S3BucketConfig())
        """
        self.conf = conf      
        # session is the resource which can be passed around
        # https://github.com/boto/boto3/issues/3197#issuecomment-1180516179
        self.session = boto3.session.Session(
            aws_access_key_id = self.conf.access_key_id,
            aws_secret_access_key = self.conf.secret_access_key
        )
        self.file_prefix = file_prefix
 
    def _get_s3_resource(self):
        # resource is a s3 endpoint 
        s3 = self.session.resource('s3', endpoint_url = self.conf.endpoint, verify=self.conf.verify_host)
        return s3
             
    def get_object_keys(self, limit_count: int=-1, input_attr_key = "key") -> map:
        """
        Return:
            map generator object
        """
        # bucket_items = []
        s3 = self._get_s3_resource()
        bucket = s3.Bucket(self.conf.bucket_name)
        # for obj in bucket.objects.all():
        
        # Filter: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/collections.html#filtering
        
        # https://stackoverflow.com/questions/45477181/apply-map-over-property-in-python
        # map object https://realpython.com/python-map-function/
        if limit_count < 0:
            # use attrgetter(input_attr_key) to filter object streams and transform the object steam to a stream of the obj.input_attr_key
            bucket_items_map = map(attrgetter(input_attr_key), 
                    bucket.objects.filter(Prefix=self.file_prefix)
                   )
        else:
            bucket_items_map = map(attrgetter(input_attr_key), 
                    bucket.objects.filter(Prefix=self.file_prefix).limit(limit_count)
                   )
        return bucket_items_map
    
    def transform_objects(self, s3_keys: map, bytesio_transformer: callable, s3_body_key = "Body", output_id_key: str = "name", output_content_key: str = "bytesio") -> map:
        """
        transforms the selected the s3 object with a given transformer
        """
        # callable https://realpython.com/python-callable-instances/
        s3 = self._get_s3_resource()
        
        bytesio_map = map(lambda x: {
            output_id_key : x,
            output_content_key : BytesIO(s3.Object(self.conf.bucket_name, x).get()[s3_body_key].read())
        }, s3_keys)
        
        return map(bytesio_transformer, bytesio_map)
       
    def upload_objects(self, data: map, key_mutater: callable = lambda x:x, input_id_key:str = "name", input_body_key: str = "content") -> map:
        s3 = self._get_s3_resource()
        bucket = s3.Bucket(self.conf.bucket_name)
        
        return map(lambda x:
            bucket.put_object(Key=key_mutater(x.get(input_id_key)), Body=x.get(input_body_key))
            ,data)


class S3PdfObjHelper(S3BucketHelper):
        """helper class for S3 PDF object specificly
        """
        class DataContract:
            name = "name"
            # reader = "reader"
            content = "content"
            pages = "pages"
            max_token_per_seg = 350
            bytesio = "bytesio"
            key_lead = "trans2en"
            key_origin_pattern = "pdf"
            key_new_pattern = "txt"
           
        def __init__(self, conf: S3AccessConf, file_prefix: str=""):
            super().__init__(conf, file_prefix)
                   
        @classmethod
        def pdf_reader_transformer(clz, input_dict: dict) -> dict:
            """
            transforms a byte stream data sequence to a PdfReader stream sequence
            """
            return {
                clz.DataContract.name : input_dict.get(clz.DataContract.name),
                clz.DataContract.content : PdfReader(input_dict.get(clz.DataContract.bytesio))
            }
        
        @classmethod
        def read_pages_transformer(clz, input_dict: dict) -> dict:
            """
            transforms a RdfReader stream sequence to a plain text stream sequence
            """
            return {
                clz.DataContract.name: input_dict.get(clz.DataContract.name),
                clz.DataContract.content: "".join([page.extract_text() for page in input_dict.get(clz.DataContract.content).pages])
            }
        
        @classmethod
        def segment_pages_transformer(clz, input_dict: dict) -> dict:
            """
            transforms a plain text stream sequence to 
            a segmented plain text stream sequence.
            """
            #  https://stackoverflow.com/questions/13673060/split-string-into-strings-by-length
            s = input_dict.get(clz.DataContract.content) # input string
            w = clz.DataContract.max_token_per_seg            
            return {
                clz.DataContract.name: input_dict.get(clz.DataContract.name),
                clz.DataContract.content: [s[i:i + w] for i in range(0, len(s), w)]
            }
        
        @classmethod
        def custom_pages_transformer_factory(clz, segment_transformer: callable) -> dict:
            """
            a factory function 
            Args:
                segment_transformer: a customer transformer function
                
            Returns:
                a transformer function to tranformer a stream of segmented pages
                using the input pages_transformer
            """
            def inner_func(input_dict: dict):
                segment_output = []
                # loop throught the page segments
                for segment in input_dict.get(clz.DataContract.content):
                    # apply the custom segment_transformer
                    segment_output.append(segment_transformer(segment))
                return {
                    clz.DataContract.name: input_dict.get(clz.DataContract.name),
                    clz.DataContract.content: ''.join(segment_output)
                }
            return inner_func
                 
        @classmethod
        def s3_key_mutater(clz, old_key: str) -> str:
            return f"{clz.DataContract.key_lead}/{old_key.replace(clz.DataContract.key_origin_pattern, clz.DataContract.key_new_pattern)}"
        


class KwargsBuilder:
    """
    merge the kwargs from the list of kwargs for the Transformer model or pipeline
    """

    def __init__(self, args_list: list = []):
        self.kwargs = {}
        self.args_list = args_list

    def override(self, kwargs: dict):
        self.args_list.append(kwargs)
        return self

    def build(self) -> dict:
        """
        during the build, merge the kwargs from the list of args_list,
        the later kwargs always override the prevous kwargs during the merge.
        """
        # https://peps.python.org/pep-0584/
        # https://stackoverflow.com/questions/38987/how-do-i-merge-two-dictionaries-in-a-single-expression-in-python
        # merge dictionary python 3.9 |= operator, python 3.8 {**d1, **d2}
        for kwargs in self.args_list:
            # kwargs override the previous self.kwargs
            self.kwargs = {**self.kwargs, **kwargs}
        return self.kwargs

    def __repr__(self) -> str:
        return self.kwargs

    def __str__(self) -> str:
        return self.__repr__()


class ModelInfo:
    def __init__(self, model_family: str, inst_msg_begin: str, inst_msg_end: str):
        self.model_family = model_family
        self.inst_msg_begin = inst_msg_begin
        self.inst_msg_end = inst_msg_end

    def __repr__(self) -> str:
        # output the raw string of the object
        # {value!r} calls the repr() method of the value to show the raw string inside the f string method
        return f"ModelInfo(model_family={self.model_family!r}, inst_msg_begin={self.inst_msg_begin!r}, inst_msg_end={self.inst_msg_end!r})"

    def __str__(self) -> str:
        return self.__repr__()


class ModelCatalog:
    """
    mistral instruction example:
    https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.2
    llama2 instruction examples:
    https://huggingface.co/TheBloke/Llama-2-7B-Chat-GGUF
    """

    META_FAMILY = "meta-llama"
    MISTRAL_FAMILY = "mistralai"
    GOOGLE_FAMILY = "google"
    INST_BEGINN_MSG_MAP = {
        MISTRAL_FAMILY: """<s>[INST] You are a helpful, respectful and honest assistant.
Always answer as helpfully as possible using the context text provided.
Your answers should only answer the question once and not have any text after the answer is done.\n
If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct.
If you don't know the answer to a question, please don't share false information. Just return \"</s>\"
""",
        META_FAMILY: """[INST]<<SYS>>You are a helpful, respectful and honest assistant.
Always answer as helpfully as possible using the context text provided.
Your answers should only answer the question once and not have any text after the answer is done.\n
If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct.
If you don't know the answer to a question, please don't share false information.<</SYS>>
""",    
        GOOGLE_FAMILY: """<start_of_turn>user
You are a helpful, respectful and honest assistant.
Always answer as helpfully as possible using the context text provided.
Your answers should only answer the question once, concise and do not have any text after the answer is done.
Your answers should only be text and not include any HTML, bullet points, or other markup.

If a question does not make sense or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information. Just say I don't know.
""",

    }
    INST_END_MSG_MAP = {
        MISTRAL_FAMILY: "[/INST]",
        META_FAMILY: "[/INST]",
        GOOGLE_FAMILY: "<end_of_turn>\n<start_of_turn>model\n",
    }

    @classmethod
    def get_model_info(cls, model_name) -> "ModelInfo":
        """
        Args:
            model_name: the model name such as mistralai/Mistral-7B-Instruct-v0.2 from huggingface
        """
        if model_name.startswith(cls.META_FAMILY):
            model_info = ModelInfo(
                cls.META_FAMILY, 
                cls.INST_BEGINN_MSG_MAP[cls.META_FAMILY], 
                cls.INST_END_MSG_MAP[cls.META_FAMILY]
            )
        elif model_name.startswith(cls.MISTRAL_FAMILY):
            model_info = ModelInfo(
                cls.MISTRAL_FAMILY,
                cls.INST_BEGINN_MSG_MAP[cls.MISTRAL_FAMILY],
                cls.INST_END_MSG_MAP[cls.MISTRAL_FAMILY]
            )
        elif model_name.startswith(cls.GOOGLE_FAMILY):
            model_info = ModelInfo(
                cls.GOOGLE_FAMILY,
                cls.INST_BEGINN_MSG_MAP[cls.GOOGLE_FAMILY],
                cls.INST_END_MSG_MAP[cls.GOOGLE_FAMILY]
            )
        else:
            model_info = ModelInfo("", "", "")
        return model_info


class PromptHelper:
    def __init__(
        self,
        model_info: ModelInfo = ModelCatalog.get_model_info(
            ModelCatalog.MISTRAL_FAMILY
        ),
    ):
        self.model_info = model_info

    def gen_prompt(self, query: str) -> str:
        if query is not None or len(query) > 0:
            prompt = f"""{self.model_info.inst_msg_begin}\n{query}\n{self.model_info.inst_msg_end}"""
        else:
            prompt = (
                f"""{self.model_info.inst_msg_begin}{self.model_info.inst_msg_end}"""
            )
        return prompt

    def get_inst_msg(self) -> str:
        return self.gen_prompt("")



import pandas as pd

def left_outer_join(df_left, df_right, on='id') -> pd.DataFrame:
    # https://stackoverflow.com/questions/50543326/how-to-do-left-outer-join-exclusion-in-pandas
    # indicator=True parameter adds a column to the output dataframe called _merge that indicates the source of each row.
    df = pd.merge(df_left, df_right, on=[on], how="left", indicator=True
              ).query('_merge=="left_only"')
    # delete the _merge column
    df = df.drop(columns=['_merge'])
    return df
    

def exp1():
    df_left = pd.DataFrame({'id': ['a', 'b', 'c', 'd']})
    df_right = pd.DataFrame({'id': ['b', 'e']})

    left_joined = left_outer_join(df_left, df_right)
    print(left_joined)


def exp2():
    df_left = pd.DataFrame({'id': ['a']})
    df_right = pd.DataFrame({'id': ['b', 'e']})

    left_joined = left_outer_join(df_left, df_right)
    print(left_joined)



if __name__ == '__main__':
    exp1()
    exp2()





import re
from langchain.output_parsers.structured import StructuredOutputParser

class StructuredOutputParserHelper:
    @classmethod
    def parse_response_str(
        clz, parser_response: str, output_parser: StructuredOutputParser, verbose: bool = False
    ) -> dict:
        # https://stackoverflow.com/questions/24667065/python-regex-difference-between-and/24667099#24667099
        # https://stackoverflow.com/questions/33312175/matching-any-character-including-newlines-in-a-python-regex-subexpression-not-g/33312193#33312193
        # (.+) is greedy, (.+?) stops at the first match
        try:
            # remove the leading and tailing text outside the ``` ```
            post_proccessed_response = re.search(
                r"```[\s\S]+```", parser_response
            ).group(0)
            # remove the comment // patient weight in kilogram \n} and keep the \n}
            # re.DOTALL is a flag that makes . match newlines as well, since default . doesn't match newline characters
            post_proccessed_response = re.sub(r"//.*}", "\n}", post_proccessed_response, flags=re.DOTALL)
            # use reg expr to check if the post_proccessed_response contains }``` string, if not add it
            #  If it finds a match, it returns a match object. If no matches are found, it returns None.
            if not re.search(r"}.*```", post_proccessed_response, flags=re.DOTALL):
                post_proccessed_response += "}\n```"

            if verbose:
                print(post_proccessed_response)
            output_dict = output_parser.parse(post_proccessed_response)
            key = output_parser.response_schemas[0].name
            if output_dict.get(key) is None:
                output_dict[key] = ""
        except Exception as e:
            print(e)
            output_dict = {}
        return output_dict

    @classmethod
    def parse_response_dict(
        clz,
        parser_response: dict,
        output_parser: StructuredOutputParser,
        text_key: str = "text",
        verbose: bool = False,
    ) -> dict:
        parser_response_str = parser_response.get(text_key, "")
        return clz.parse_response_str(
            parser_response=parser_response_str,
            output_parser=output_parser,
            verbose=verbose,
        )



import torch
import os
import subprocess
import re
import sys
import time
from dataclasses import dataclass


@dataclass
class DirectorySetting:
    """set the directory for the model download"""
    home_dir: str="/home/jovyan/llm-models"
    transformers_cache_home: str="core-kind/yinwang/models"
    huggingface_token_file: str="core-kind/yinwang/.cache/huggingface/token"

    def get_cache_home(self):
        """get the cache home"""
        return f"{self.home_dir}/{self.transformers_cache_home}"
    
    def get_token_file(self):
        """get the token file"""
        return f"{self.home_dir}/{self.huggingface_token_file}"


DIR_MODE_MAP = {
    "kf_notebook": DirectorySetting(),
    "mac_local": DirectorySetting(home_dir="/Users/yingding", transformers_cache_home="MODELS", huggingface_token_file="MODELS/.huggingface_token"),
}


class TokenHelper():
    @staticmethod
    def need_token(model_type: str, model_name_prefix: str="llama"):
        """check if the model needs token"""
        return model_type.startswith(model_name_prefix)
       
    @staticmethod
    def get_token(dir_setting: DirectorySetting):
        """get the token from the token file"""
        token_file_path = dir_setting.get_token_file()
        with open(token_file_path, "r") as file:
            # file read add a new line to the token, remove it.
            token = file.read().replace('\n', '')
        return token
    
    @staticmethod
    def gen_token_kwargs(model_type: str, dir_setting: DirectorySetting):
        """
        """
        if TokenHelper.need_token(model_type):
            # kwargs = {"use_auth_token": get_token(dir_setting)}
            token_kwargs = {
                "token": TokenHelper.get_token(dir_setting),
                # "truncation_side": "left",
                # "return_tensors": "pt",            
                            }
            print("huggingface token loaded")
        else:
            token_kwargs = {}
            print("huggingface token is NOT needed")
        return token_kwargs


class AcceleratorHelper():
    @staticmethod
    def print_container_info() -> None:
        print("-"*10)
        print(time.strftime("%Y-%m-%d_%H-%M-%S"))
        print(f"python version: {sys.version}")
        print(f"torch version: {torch.__version__}")
        print("-"*10)

    @staticmethod
    def nvidia_device_info() -> str:
        """get the nvidia MIGs device uuid and GPU uuid 
        """
        # blocking call
        result = subprocess.run(["nvidia-smi", "-L"], stdout=subprocess.PIPE)
        # decode the byte object, returns string with \n
        cmd_out_str = result.stdout.decode('utf-8')
        return [line.strip() for line in cmd_out_str.split('\n') if len(line) > 0]

    @staticmethod
    def extract_nvidia_device_uuids(input: str):
        """parse the nvidia devices uuid from the nvidia device info str
        """
        try:
            # r'' before the search pattern indicates it is a raw string, 
            # otherwise "" instead of single quote
            uuid = re.search(r'UUID\:\s(.+?)\)', input).group(1)
        except AttributeError:
            # "UUID\:\s" and "\)" not found
            uuid = ""
        return uuid

    @staticmethod
    def nvidia_device_uuids_filtered_by(is_mig: bool = False, log_output: bool = False) -> str:
        """get a comma separated str of nvidia MIGs devices
        """
        info_list = AcceleratorHelper.nvidia_device_info()
        if is_mig:
            # skip the first GPU ID, get the MIGs IDS
            uuid_list = [AcceleratorHelper.extract_nvidia_device_uuids(e) for e in info_list[1:]]
        else: # all GPU devices
            uuid_list = [AcceleratorHelper.extract_nvidia_device_uuids(e) for e in info_list]
        if log_output is not None and log_output:
            print(uuid_list)
        
        # if multi gpus need to join the device together for pytorch
        return ",".join(uuid_list)

    @staticmethod
    # def init_cuda_torch(uuids: str, data_path: str, debug: bool = False) -> None:
    def init_cuda_torch(uuids: str, dir_setting: DirectorySetting, debug: bool = False) -> None:
        """setup the default env variables for transformers
        
        Args:
          uuids: a comma separate str of nvidia gpu/mig uuids
        """
        os.environ["WORLD_SIZE"] = "1" 
        os.environ["CUDA_VISIBLE_DEVICES"] = uuids 
        os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:512" #512
        os.environ['XDG_CACHE_HOME'] = dir_setting.get_cache_home()
        # os.environ['XDG_CACHE_HOME'] = f"{data_path}/models"
        if (debug):
            # for debugging      
            os.environ["CUDA_LAUNCH_BLOCKING"] = "1" 
        # os.environ["TOKENIZERS_PARALLELISM"]="false"
            
    @staticmethod
    def init_mps_torch(dir_setting: DirectorySetting) -> None:
        """setup the default env variables for transformers
        """
        os.environ["WORLD_SIZE"] = "1" 
        os.environ['XDG_CACHE_HOME'] = dir_setting.get_cache_home()
        # os.environ['TOKENIZERS_PARALLELISM'] = "false"
        # https://github.com/huggingface/transformers/issues/5486?ref=assemblyai.com


class AcceleratorStatus():
    # Reference: https://stackoverflow.com/questions/58216000/get-total-amount-of-free-gpu-memory-and-available-using-pytorch
    # from typing import Tuple
    def byte_gb_info(self, byte_mem) -> str:
        """calculate the byte size to GB size for better human readable"""
        # format the f string float with :.2f to decimal digits
        # https://zetcode.com/python/fstring/
        return f"{(byte_mem/1024**3):4f} GB"
    
    def gpu_usage(self) -> None:
        pass

    # factory method to create the correct accelerator status class
    @staticmethod
    def create_accelerator_status() -> 'AcceleratorStatus':
        # check if the MPS is enabled
        if torch.backends.mps.is_available():
            return MpsAcceleratorStatus()
        elif torch.cuda.is_available():
            return CudaAcceleratorStatus()
        else:
            return AcceleratorStatus()

        
class MpsAcceleratorStatus(AcceleratorStatus):
    def __init__(self):
        pass

    def accelerator_mem_info(self):
        # allocated
        a = torch.mps.driver_allocated_memory()
        print(f"Allocated memory : {self.byte_gb_info(a)}")

    def gpu_usage(self) -> None:
        print("-"*20)            
        self.accelerator_mem_info()
        print("-"*20)


class CudaAcceleratorStatus(AcceleratorStatus):
    def __init__(self):
        pass

    def accelerator_mem_info(self, device_idx: int):
        # total
        t = torch.cuda.get_device_properties(device_idx).total_memory
        # usable
        r = torch.cuda.memory_reserved(device_idx)
        # allocated
        a = torch.cuda.memory_allocated(device_idx)
        # still free
        f = r-a
        # unit = "GB"   
        print( # "GPU memory info:\n" + 
              f"Physical  memory : {self.byte_gb_info(t)}\n" + 
              f"Reserved  memory : {self.byte_gb_info(r)}\n" + 
              f"Allocated memory : {self.byte_gb_info(a)}\n" + 
              f"Free      memory : {self.byte_gb_info(f)}")

    def accelerator_compute_info(self, device_idx: int) -> None:
        name = torch.cuda.get_device_properties(device_idx).name
        count = torch.cuda.get_device_properties(device_idx).multi_processor_count
        print(f"Device name      : {name} \n" +
              f"Device idx       : {device_idx} \n" +
              f"No. of processors: {count}")    

    def gpu_usage(self) -> None:        
        num_of_gpus = torch.cuda.device_count();
        # this shows only the gpu device, not the MIG
        print(f"num_of_gpus: {num_of_gpus}")
        for device_idx in range(torch.cuda.device_count()):
            print("-"*20)
            self.accelerator_compute_info(device_idx)                 
            self.accelerator_mem_info(device_idx)
            print("-"*20)
        